# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Gazoo Device Manager (gdm) is a single interface for device interactions.

It manages communication and setup of smart devices to simplify daily tasks
(software updating a unit, getting logs) as well as serving as a device driver
for automated (python) tests.

Areas of focus: ease of use (including ease of installation, setup, updating),
robustness, scalability.

Examples of functionality:
* Detecting and setting up devices when they are attached (create config
  files automatically).
* Setting up shared permissions/file locations so all users can access test
  hardware (if they can log onto the pc).
* Sharing a single device UART for logging and commands, to allow logs to be
  streamed "live" during a test while still sending commands/responses during a
  test.

Command Line Interface (CLI):
There two groups of commands.
"Manager" commands operate on local config files on the pc.
  The command may talk to devices (detect), but they don't change the
  state of the device (i.e. read-only).
  The command may change the config files stored on the pc, to detect
  new devices, add an alias, etc.
  For manager commands, the format is
    "gdm <command> <device> [<additional arguments>]".
    For example: "gdm log cambrionix-1234".
"Device" commands talk to or configure connected devices.
  This includes upgrade, reboot, etc.
  For device commands, the format is
    "gdm issue <device> - <command> [<arguments>]".
  For example: "gdm issue cambrionix-1234 - reboot".

You can get more details on a particular command at the command line with
"gdm man DEVICENAME COMMAND" (does not require a connected device) or
"gdm issue DEVICENAME - COMMAND -- --help" (requires a connected device).
For example: "gdm issue cambrionix-1234 - reboot -- --help"
             "gdm man cambrionix reboot"

Refer to https://github.com/google/gazoo-device for full documentation.
"""
import gc
import logging
import multiprocessing
import os
import signal
import sys

from gazoo_device import _version
from gazoo_device import gazoo_device_controllers
from gazoo_device import gdm_logger
from gazoo_device import manager
from gazoo_device import mobly_controller
from gazoo_device import package_registrar
from gazoo_device.utility import common_utils

Manager = manager.Manager
register = package_registrar.register
version = _version.version
__version__ = _version.version

# For Mobly integration
MOBLY_CONTROLLER_CONFIG_NAME = "GazooDevice"
create = mobly_controller.create
destroy = mobly_controller.destroy
get_info = mobly_controller.get_info

# Defend against inadvertent basicConfig, which adds log noise
logging.getLogger().addHandler(logging.NullHandler())


# Ensure that atexit handlers run when killed by SIGTERM
def graceful_exit(*args, **kwargs):  # pylint: disable=unused-argument
  raise SystemExit(0)


signal.signal(signal.SIGTERM, graceful_exit)

if sys.platform == "darwin" and sys.version_info >= (3, 8):
  # Workaround for b/160958582: Python >= 3.8 defaults to "spawn" start method
  # on Macs. GDM isn't compatible with "spawn" yet, so use the "fork" method
  # instead.
  multiprocessing.set_start_method("fork", force=True)

# Set up logger
gdm_logger.initialize_logger()


def _after_fork():
  """Re-enables garbage collection in both parent & child process."""
  gc.enable()


def _before_fork():
  """Collects garbage, disables periodic GC, and flushes logger messages.

  This ensures that the logging thread is not logging anything during os.fork()
  calls. If the logging thread is logging during an os.fork() call, child
  processes may be forked with an acquired stdout lock, which will cause a
  deadlock when the child process finishes. Child processes attempt to flush the
  stdout buffer before exiting and will hang indefinitely if forked with an
  acquired stdout buffer lock.
  """
  gc.disable()
  gc.collect()
  gdm_logger.flush_queue_messages()


# Periodic GC is active in child processes created via os.fork.
# Trigger a collection in the parent process prior to forking to prevent
# parent process's garbage from being copied over to the child and later being
# collected in each process. See b/150018610.
common_utils.register_at_fork(before=_before_fork,
                              after_in_parent=_after_fork,
                              after_in_child=_after_fork)

# Register device classes and capabilities built into GDM
package_registrar.register(gazoo_device_controllers)



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Common types used across the gazoo_device module and externally."""
from typing import Dict, Union

from gazoo_device.base_classes import auxiliary_device
from gazoo_device.base_classes import gazoo_device_base
from typing_extensions import Literal  # Moved to typing in 3.8+

# A device object, as returned by Manager.create_device().
Device = Union[auxiliary_device.AuxiliaryDevice,
               gazoo_device_base.GazooDeviceBase]

PropertyName = str
PropertyValue = Union[bool, float, int, str, None]
DeviceConfig = Dict[PropertyName, PropertyValue]

DeviceName = str
PersistentConfigsDict = Dict[
    Literal["devices", "other_devices"],
    Dict[DeviceName, DeviceConfig]]
OptionalConfigsDict = Dict[
    Literal["device_options", "other_device_options"],
    Dict[DeviceName, DeviceConfig]]
# Such entries are present in Manager._devices and Manager.other_devices.
ManagerDeviceConfigDict = Dict[Literal["persistent", "options"],
                               DeviceConfig]
# Returned by <device_instance>.props.
DeviceClassConfigDict = Dict[Literal["persistent_identifiers", "optional"],
                             DeviceConfig]



# Copyright 2021 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Default implementation of the Pigweed RPC device common capability."""
import time
from typing import Any, Callable, Optional
from gazoo_device import decorators
from gazoo_device import errors
from gazoo_device import gdm_logger
from gazoo_device.capabilities.interfaces import pwrpc_common_base
from gazoo_device.switchboard.transports import pigweed_rpc_transport
try:
  # pylint: disable=g-import-not-at-top
  # pytype: disable=import-error
  from device_service import device_service_pb2
  # pytype: enable=import-error
except ImportError:
  device_service_pb2 = None

logger = gdm_logger.get_logger()
_RPC_TIMEOUT = 10


class PwRPCCommonDefault(pwrpc_common_base.PwRPCCommonBase):
  """Pigweed RPC common capability for devices communicating over PwRPC."""

  def __init__(self,
               device_name: str,
               switchboard_call: Callable[..., Any],
               switchboard_call_expect: Callable[..., Any]):
    """Create an instance of the PwRPCCommonDefault capability.

    Args:
      device_name: Device name used for logging.
      switchboard_call: The switchboard.call method which calls to the endpoint.
        See more examples in nrf_pigweed_lighting.py.
      switchboard_call_expect: The switchboard.call_and_expect method.
    """
    super().__init__(device_name=device_name)
    self._switchboard_call = switchboard_call
    self._switchboard_call_expect = switchboard_call_expect

  @decorators.DynamicProperty
  def vendor_id(self) -> str:
    """The vendor id of the device."""
    return self._get_static_info("vendor_id")

  @decorators.DynamicProperty
  def product_id(self) -> str:
    """The product id of the device."""
    return self._get_static_info("product_id")

  @decorators.DynamicProperty
  def software_version(self) -> str:
    """The software version of the device."""
    return self._get_static_info("software_version")

  @decorators.CapabilityLogDecorator(logger)
  def reboot(self,
             no_wait: bool = False,
             rpc_timeout_s: int = _RPC_TIMEOUT,
             bootup_logline_regex: Optional[str] = None,
             bootup_timeout: Optional[int] = None):
    """Reboots the device.

    Args:
      no_wait: Return before reboot completes.
      rpc_timeout_s: Timeout (s) for RPC call.
      bootup_logline_regex: Device logline indicating booting up, not applicable
        if no_wait is True.
      bootup_timeout: Timeout (s) to wait for the bootup message, not applicable
        if no_wait is True.
    """
    if no_wait:
      self._trigger_device_action(action="Reboot",
                                  rpc_timeout_s=rpc_timeout_s)
    else:
      self._trigger_device_action(action="Reboot",
                                  rpc_timeout_s=rpc_timeout_s,
                                  expect_regex=bootup_logline_regex,
                                  expect_timeout_s=bootup_timeout)
      self._wait_for_bootup_complete(bootup_timeout)

  @decorators.CapabilityLogDecorator(logger)
  def factory_reset(self,
                    no_wait: bool = False,
                    rpc_timeout_s: int = _RPC_TIMEOUT,
                    bootup_logline_regex: Optional[str] = None,
                    bootup_timeout: Optional[int] = None):
    """Factory resets the device.

    Args:
      no_wait: Return before factory-reset completes.
      rpc_timeout_s: Timeout (s) for RPC call.
      bootup_logline_regex: Device logline indicating booting up, not applicable
      if no_wait is True.
      bootup_timeout: Timeout (s) to wait for the bootup message, not applicable
      if no_wait is True.
    """
    if no_wait:
      self._trigger_device_action(action="FactoryReset",
                                  rpc_timeout_s=rpc_timeout_s)
    else:
      self._trigger_device_action(action="FactoryReset",
                                  rpc_timeout_s=rpc_timeout_s,
                                  expect_regex=bootup_logline_regex,
                                  expect_timeout_s=bootup_timeout)
      self._wait_for_bootup_complete(bootup_timeout)

  @decorators.CapabilityLogDecorator(logger)
  def ota(self):
    """Triggers OTA to the device."""
    self._trigger_device_action("TriggerOta")

  def _get_static_info(self, property_name: str) -> str:
    """Returns device static information.

    Args:
      property_name: Static property name which currently supports: vendor_id,
        product_id and software_version.

    Raises:
      DeviceError: The ack status is not true or an invalid property_name is
      given.
    """
    ack, payload_in_bytes = self._switchboard_call(
        method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
        method_args=("Device", "GetDeviceInfo"),
        method_kwargs={})
    if not ack:
      raise errors.DeviceError(
          f"{self._device_name} getting static info failed.")
    payload = device_service_pb2.DeviceInfo.FromString(payload_in_bytes)
    device_property = getattr(payload, property_name, None)
    if device_property is None:
      raise errors.DeviceError(f"{property_name} doesn't exist in static info.")
    return device_property

  def _trigger_device_action(
      self,
      action: str,
      rpc_timeout_s: int = _RPC_TIMEOUT,
      expect_regex: Optional[str] = None,
      expect_timeout_s: Optional[int] = None) -> None:
    """Triggers specific device action.

    Args:
      action: Device actions including reboot, factory-reset and OTA.
      rpc_timeout_s: Timeout (s) for RPC call.
      expect_regex: Expected device logline regex.
      expect_timeout_s: Timeout (s) to wait for the expected regex.

    Raises:
      DeviceError: The ack status is not true.
    """
    if expect_regex is None and expect_timeout_s is None:
      ack, _ = self._switchboard_call(
          method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
          method_args=("Device", action),
          method_kwargs={"pw_rpc_timeout_s": rpc_timeout_s})
    elif expect_regex is not None and expect_timeout_s is not None:
      _, (ack, _) = self._switchboard_call_expect(
          method=pigweed_rpc_transport.PigweedRPCTransport.rpc,
          pattern_list=[expect_regex],
          timeout=expect_timeout_s,
          method_args=("Device", action),
          method_kwargs={"pw_rpc_timeout_s": rpc_timeout_s},
          raise_for_timeout=True)
    else:
      raise ValueError("Only one of \"expect_regex\", \"expect_timeout_s\" "
                       "arguments was provided. Both or neither should be "
                       "provided.")
    if not ack:
      raise errors.DeviceError(f"{self._device_name} triggering {action} failed"
                               ": The action did not succeed")

  def _wait_for_bootup_complete(self, bootup_timeout: int):
    """Waits for device to boot up.

    Args:
      bootup_timeout: Max time to wait for bootup to complete (in seconds).

    Raises:
      DeviceError: If device did not boot up successfully in given timeout.
    """
    start_time = time.time()
    bootup_time = start_time + bootup_timeout
    while time.time() < bootup_time:
      try:
        logger.debug(f"{self._device_name} responded to the RPC call: "
                     f"{self.software_version}")
        logger.info(f"{self._device_name} booted up in "
                    f"{time.time() - start_time}s")
        return
      except errors.DeviceError:
        logger.debug(f"{self._device_name} hasn't booted up yet.")
      time.sleep(0.5)
    raise errors.DeviceError(f"Failed to boot up within {bootup_timeout}s.")
